#!/usr/bin/env python
# vim: set expandtab tabstop=4 shiftwidth=4:
#
# UQM Starmap Viewer
# Copyright (C) 2009 CJ Kucera
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
import sys
import math
import string
import os.path
import colorsys
import cairo
import pango
import pangocairo
import pygtk
import gtk
import gtk.glade
import gobject
import cPickle
pygtk.require('2.0')

# Some globals.  This is... dumb?  Alas.
# Note that especially the max_width and max_height probably aren't really portable across
# users' themes...
app_name = 'UQM Starmap Viewer'
app_version = '1.0.0'
app_url = 'http://apocalyptech.com/url/'
min_vals = [ 'common', 'corrosive', 'base', 'noble', 'rare', 'precious', 'radioactive', 'exotic' ]
max_presets = 7
max_width = 1207
max_height = 1053
init_width = 1000
init_height = 700
quasi_dim = 150
quasi_offset = 425
quasi_line_offset = -25

class MinData():
    """
    A class to hold mineral data.  Each planet has one of these, and each
    system uses a few of these for aggregate data as well.
    """
    def __init__(self, common=0, corrosive=0, base=0, noble=0, rare=0, precious=0, radioactive=0, exotic=0):
        self.common = common
        self.corrosive = corrosive
        self.base = base
        self.noble = noble
        self.rare = rare
        self.precious = precious
        self.radioactive = radioactive
        self.exotic = exotic

    def value(self):
        """
        Returns the RU value for the given set of minerals
        """
        return self.common+(2*self.corrosive)+(3*self.base)+(4*self.noble)+ \
            (5*self.rare)+(6*self.precious)+(8*self.radioactive)+(25*self.exotic)

    def weight(self):
        """
        Returns the weight of all the minerals (that is, how much room they'll
        consume in your cargo holds)
        """
        return self.common+self.corrosive+self.base+self.noble+self.rare+self.precious+self.radioactive+self.exotic

    def worth(self):
        """
        This is labeled "MVpU" or "Mineral Value per Unit" in the GUI typically -
        it's just a measure of how much bang for the buck you get.  A planet with
        7k RU in Exotic Minerals has more "worth" than one with 7k RU in Base Metals
        because it'll take far less room in your cargo holds.
        """
        if (self.weight() != 0):
            return self.value()/float(self.weight())
        else:
            return 0

class Planet():
    """
    Just some basic data about a planet.  Nothing terribly exciting.
    """
    def __init__(self, id, name, type, tectonics, weather, temp, gravity, bio, bio_danger, mineral):
        self.id = id
        self.name = name
        self.type = type
        self.tectonics = tectonics
        self.weather = weather
        self.temp = temp
        self.gravity = gravity
        self.bio = bio
        self.bio_danger = bio_danger
        self.mineral = mineral

class Filter():
    """
    A class to hold filters that we'll use to limit systems/planets.
    """
    def __init__(self):
        """
        Initial filter doesn't actually contain anything.
        """
        self.reset()

    def reset(self):
        """
        Resets the individual filters to None
        """
        self.filters = []

    def add(self, filter):
        """
        Adds a new filter to the filter chain.
        """
        self.filters.append(filter)

    def filtering(self):
        """
        Returns whether or not we have an active filter.
        """
        return (len(self.filters) > 0)

    def approve(self, object):
        """
        Returns true if the given object is approved, false if denied.
        """
        approved = True
        for filter in self.filters:
            if (not filter.approve(object)):
                approved = False
                break
        return approved

class ProxFilter():
    """
    A class to hold filter information related to proximity.
    """
    def __init__(self, system, radius):
        self.system = system
        self.radius = radius

    def approve(self, planet):
        return (self.system and planet.distance_to(self.system) <= self.radius)

class TypeFilter():
    """
    A class to hold filter information about what kinds of planets we'll accept.
    Note that we'll be getting passed a system to approve(), so we must loop through
    the planets ourselves.
    """
    def __init__(self, type):
        self.type = type
        self.typelen = len(self.type)

    def approve(self, system):
        """
        Returns true if the system contains a planet of the given type,
        or false if not.
        """
        for planet in system.planets:
            if (planet.type[:self.typelen] == self.type):
                return True
        return False

class NameFilter():
    """
    A class to hold filter information which matches system names.  Note that it'll
    also search in the 'extra' field if the 'special info' box is checked.
    """
    def __init__(self, name, specialchecked):
        self.name = name.lower()
        self.specialchecked = specialchecked

    def approve(self, system):
        """
        Returns true if the system matches the given name, false if not.
        """
        return (system.fullname.lower().find(self.name) > -1 or
                (self.specialchecked and system.extra.lower().find(self.name) > -1))

class ConstFilter():
    """
    A class to hold filter information which matches constellation names.
    """
    def __init__(self, name):
        self.name = name

    def approve(self, system):
        """
        Returns true if the system matches the given constellation name, false if not.
        """
        return (system.name == self.name)

class SafetyFilter():
    """
    A class to hold safety filter information, used to make the processing of
    mineral and bio data aggregates easier.
    """
    def __init__(self):
        """
        Initial filter will end up accepting all planets on the map
        """
        self.tectonics_dir = -1
        self.weather_dir = -1
        self.temp_dir = -1
        self.bio_dir = -1
        self.tectonics_val = 8
        self.weather_val = 8
        self.temp_val = 5200
        self.bio_val = 400
        self.activate()

    def activate(self):
        """
        If any of the internal "direction" variables (*_dir) change, call this
        to make sure that the testing functions are updated properly.
        """

        # Define the function to test tectonics
        if (self.tectonics_dir < 0):
            self.tectonics = self.tectonics_lte
        else:
            self.tectonics = self.tectonics_gte

        # Define the function to test weather
        if (self.weather_dir < 0):
            self.weather = self.weather_lte
        else:
            self.weather = self.weather_gte

        # Define the function to test temp
        if (self.temp_dir < 0):
            self.temp = self.temp_lte
        else:
            self.temp = self.temp_gte

        # Define the function to test bio
        if (self.bio_dir < 0):
            self.bio = self.bio_lte
        else:
            self.bio = self.bio_gte

    def approve(self, planet):
        """
        Returns true if the given planet is approved, false if denied.
        """
        return (self.tectonics(planet) and self.weather(planet) and self.temp(planet) and self.bio(planet))

    # The various functions for testing follow
    def tectonics_lte(self, planet):
        return (planet.tectonics <= self.tectonics_val)
    def tectonics_gte(self, planet):
        return (planet.tectonics >= self.tectonics_val)
    def weather_lte(self, planet):
        return (planet.weather <= self.weather_val)
    def weather_gte(self, planet):
        return (planet.weather >= self.weather_val)
    def temp_lte(self, planet):
        return (planet.temp <= self.temp_val)
    def temp_gte(self, planet):
        return (planet.temp >= self.temp_val)
    def bio_lte(self, planet):
        return (planet.bio_danger <= self.bio_val)
    def bio_gte(self, planet):
        return (planet.bio_danger >= self.bio_val)

class System():
    """
    Class to hold information about a specific system.  Most of the "fun"
    data in here is the aggregate data.  We keep track of two sets of
    aggregates; one which holds the total value of the system, and one
    which holds the value which might be modified by safety filters.
    """
    def __init__(self, id, name, position, x, y, stype, extra):
        self.quasi = False
        self.id = id
        self.name = name
        self.x = x
        self.y = y
        self.draw_x = int(x/10)
        self.draw_y = 1000-int(y/10)
        self.position = position
        self.stype = stype
        self.extra = extra
        self.highlight = True
        self.planets = []
        self.min_agg = None
        self.min_agg_full = None
        self.bio_agg = None
        self.bio_agg_full = None
        self.bio_danger_agg = None
        self.bio_danger_agg_full = None
        if (self.position and self.position != ''):
            self.fullname = '%s %s' % (self.position, self.name)
        else:
            self.fullname = self.name
        self.quasispace = False

    def addplanet(self, planet):
        """
        Adds a planet to the system; only used when initially importing the data
        """
        self.planets.append(planet)

    def distance_to(self, system):
        """
        Returns the distance between this system and the specified one
        """
        if system:
            return (math.sqrt((self.x-system.x)**2+(self.y-system.y)**2)/10)

    def set_aggregate(self, dispfilter, aggfilter):
        """
        Sets the aggregate mineral/bio value of the planet, using the given
        filters.  Will set the "highlight" property appropriately.
        """
        self.highlight = dispfilter.approve(self)
        self.min_agg = MinData()
        self.min_agg_full = MinData()
        self.bio_agg = 0
        self.bio_agg_full = 0
        self.bio_danger_agg = 0
        self.bio_danger_agg_full = 0
        for planet in self.planets:
            self.bio_agg_full += planet.bio
            self.bio_danger_agg_full += planet.bio_danger
            for min in min_vals:
                self.min_agg_full.__dict__[min] += planet.mineral.__dict__[min]
            if (aggfilter.approve(planet)):
                self.bio_agg += planet.bio
                self.bio_danger_agg += planet.bio_danger
                for min in min_vals:
                    self.min_agg.__dict__[min] += planet.mineral.__dict__[min]
        return (self.min_agg.value(), self.bio_agg)

    # Aggregate reporting follows;  I suppose this isn't very Pythonic, but
    # it means slightly less typing later on.  Go figure.
    def agg_value(self):
        return self.min_agg.value()
    def agg_weight(self):
        return self.min_agg.weight()
    def agg_worth(self):
        return self.min_agg.worth()
    def full_agg_value(self):
        return self.min_agg_full.value()
    def full_agg_weight(self):
        return self.min_agg_full.weight()
    def full_agg_worth(self):
        return self.min_agg_full.worth()
    def bio_agg_value(self):
        return self.bio_agg
    def full_bio_agg_value(self):
        return self.bio_agg_full
    def bio_danger_agg_value(self):
        return self.bio_danger_agg
    def full_bio_danger_agg_value(self):
        return self.bio_danger_agg_full

class Systems():
    """
    Aggregate class to hold all of our systems.

    Note that we set up two global filters - one to filter aggregate data based
    on planets (ie: the saftey filter), and the other to filter highlighting of
    systems (ie: proximity, etc).  By default these will be empty, and approve
    all systems/planets.
    """
    def __init__(self):
        self.dispfilter = Filter()
        self.aggfilter = Filter()
        self.systems = {}
        self.coords = []
        self.agg_min_value = 9999
        self.agg_max_value = 0
        self.agg_spread = 0
        self.bio_agg_min_value = 9999
        self.bio_agg_max_value = 0
        self.bio_spread = 0
        self.connections = []
        self.quasispace = []
        for x in range(1001):
            self.coords.append([])
            for y in range(1001):
                self.coords[x].append(None)

    def add(self, id, name, position, x, y, stype, extra):
        """
        Add a new system, only used during the initial import.
        """
        self.systems[id] = System(id, name, position, x, y, stype, extra)
        self.coords[self.systems[id].draw_x][self.systems[id].draw_y] = self.systems[id]

    def add_quasi(self, x, y, qs_x, qs_y, label):
        """
        Adds a new quasispace exit, only used during the initial import.
        """
        self.systems[label] = Quasispace(x, y, qs_x, qs_y, label)
        self.coords[self.systems[label].draw_x][self.systems[label].draw_y] = self.systems[label]
        self.quasispace.append(self.systems[label])

    def get(self, id):
        """
        Returns a system given its ID.
        """
        try:
            return self.systems[id]
        except KeyError:
            return None

    def lookup(self, x, y):
        """
        Looks up a system given an (x,y) coord from the map
        """
        return self.coords[int(x)][int(y)]

    def getall(self):
        """
        Returns a list of all of the systems we know about
        """
        return self.systems.values()

    def process_aggregates(self):
        """
        Loop through all of our systems and calculate its aggregate values, using the
        given filters.  Will also set the "highlight" var for each system
        accordingly.  Will only calculate min/max values for planets which
        are set to be highlighted (so that the color spread will be accurate)
        """
        self.agg_min_value = 9999
        self.agg_max_value = 0
        self.agg_spread = 0
        self.bio_agg_min_value = 9999
        self.bio_agg_max_value = 0
        self.bio_spread = 0
        for system in self.getall():
            if (system.quasi):
                continue
            (min, bio) = system.set_aggregate(self.dispfilter, self.aggfilter)
            if (system.highlight):
                if (min < self.agg_min_value):
                    self.agg_min_value = min
                if (min > self.agg_max_value):
                    self.agg_max_value = min
                if (bio < self.bio_agg_min_value):
                    self.bio_agg_min_value = bio
                if (bio > self.bio_agg_max_value):
                    self.bio_agg_max_value = bio
        self.agg_spread = float(self.agg_max_value - self.agg_min_value)
        self.bio_spread = float(self.bio_agg_max_value - self.bio_agg_min_value)

    def ret_intensity(self, system):
        """
        Returns the mineral intensity of a given system, based on the calculated aggregates.
        """
        if (self.agg_spread == 0):
            return 100
        else:
            return (system.agg_value()-self.agg_min_value)/self.agg_spread

    def ret_bio_intensity(self, system):
        """
        Returns the biological intensity of a given system, based on the calculated aggregates.
        """
        if (self.bio_spread == 0):
            return 100
        else:
            return (system.bio_agg_value()-self.bio_agg_min_value)/self.bio_spread

class Quasispace():
    """
    Just a little class to hold quasispace exit information.  Not much more
    than a glorified dict, but this way we access all the data in the same
    way, and we can pretend to be a system.
    """
    def __init__(self, x, y, qs_x, qs_y, label):
        self.x = x
        self.y = y
        self.draw_x = int(x/10)
        self.draw_y = 1000-int(y/10)
        self.qs_x = qs_x
        self.qs_y = qs_y
        self.draw_quasi_x = qs_x-quasi_offset
        self.draw_quasi_y = (1000-qs_y)-quasi_offset
        self.id = label
        self.quasi = True
        self.fullname = 'Quasispace Exit %s' % (label)

    def distance_to(self, system):
        """
        Returns the distance between this system and the specified one
        """
        if system:
            return (math.sqrt((self.x-system.x)**2+(self.y-system.y)**2)/10)

class UQMMap():

    def __init__(self):
        """
        The main application class.   Internally this is sort of a mess, since it
        was build pretty piecemeal.
        """

        # First we load in our data
        try:
            df = open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data.pickle'), 'r')
            data = cPickle.load(df)
            df.close()
        except:
            print "Couldn't open data.pickle, aborting..."
            return 1

        planet_types = {}
        const_names = {}

        # Create our systems
        self.systems = Systems()
        for system in data['systems']:
            self.systems.add(system['sid'], system['name'], system['position'], system['x'], system['y'], system['stype'], system['extra'])
            const_names[system['name']] = True

        # Make sure we store our list of quasispace exits, too
        for quasi in data['quasispace']:
            self.systems.add_quasi(quasi['x'], quasi['y'], quasi['qs_x'], quasi['qs_y'], quasi['label'])

        # ... and now a list of planets
        for planet in data['planets']:
            self.systems.get(planet['sid']).addplanet(Planet(
                    planet['pid'], planet['pname'], planet['ptype'], planet['tectonics'], planet['weather'], planet['temp'], planet['gravity'],
                    planet['bio'], planet['bio_danger'],
                    MinData(planet['min_common'], planet['min_corrosive'], planet['min_base'], planet['min_noble'],
                        planet['min_rare'], planet['min_precious'], planet['min_radio'], planet['min_exotic'])
                    )
                )
            if (planet['ptype'][-6:] == ' World'):
                planet_types[planet['ptype'][:-6]] = True
            else:
                planet_types[planet['ptype']] = True

        # Let's process our constellation connection information too.
        for (sid, lids) in data['constellations'].items():
            ssystem = self.systems.systems[sid]
            ssys_x = ssystem.draw_x
            ssys_y = ssystem.draw_y
            for lid in lids:
                lsystem = self.systems.systems[lid]
                self.systems.connections.append(((ssys_x, ssys_y), (lsystem.draw_x, lsystem.draw_y)))

        # ... this should happen automatically, but regardless:
        data = None

        # Run through aggregates and calc min/max
        self.systems.process_aggregates()

        # Graphic elements we want
        self.gladefile = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'main.glade')
        self.wtree = gtk.glade.XML(self.gladefile)
        self.window = self.wtree.get_widget('mainwindow')
        self.mainscroll = self.wtree.get_widget('mainscroll')
        self.mainarea = self.wtree.get_widget('mainarea')
        self.infolabel = self.wtree.get_widget('infolabel')
        self.systeminfobox = self.wtree.get_widget('systeminfobox')
        self.systemwindow = self.wtree.get_widget('systemwindow')
        self.systemdetails_button = self.wtree.get_widget('systemdetails_button')
        self.aboutwindow = None

        # Info Labels
        self.namelabel = self.wtree.get_widget('namelabel')
        self.coordlabel = self.wtree.get_widget('coordlabel')
        self.minerallabel = self.wtree.get_widget('minerallabel')

        # Filter Controls
        self.special_check = self.wtree.get_widget('special_check')
        self.dispspecial_check = self.wtree.get_widget('dispspecial_check')
        self.quasi_check = self.wtree.get_widget('quasi_check')
        self.dispconst_check = self.wtree.get_widget('dispconst_check')
        self.visvalue_check = self.wtree.get_widget('visvalue_check')
        self.visvalue_mineral_radio = self.wtree.get_widget('visvalue_mineral_radio')
        self.visvalue_bio_radio = self.wtree.get_widget('visvalue_bio_radio')
        self.popup_check = self.wtree.get_widget('popup_check')
        self.proximity_check = self.wtree.get_widget('proximity_check')
        self.proximity_spin = self.wtree.get_widget('proximity_spin')
        self.safety_check = self.wtree.get_widget('safety_check')
        self.safety_preset_box = self.wtree.get_widget('safety_preset_box')
        self.tectonics_dir = self.wtree.get_widget('tectonics_dir')
        self.weather_dir = self.wtree.get_widget('weather_dir')
        self.temp_dir = self.wtree.get_widget('temp_dir')
        self.bio_dir = self.wtree.get_widget('bio_dir')
        self.tectonics_spin = self.wtree.get_widget('tectonics_spin')
        self.weather_spin = self.wtree.get_widget('weather_spin')
        self.temp_spin = self.wtree.get_widget('temp_spin')
        self.bio_spin = self.wtree.get_widget('bio_spin')
        self.type_check = self.wtree.get_widget('type_check')
        self.type_dropdown = self.wtree.get_widget('type_dropdown')
        self.const_check = self.wtree.get_widget('const_check')
        self.const_dropdown = self.wtree.get_widget('const_dropdown')
        self.name_check = self.wtree.get_widget('name_check')
        self.name_entry = self.wtree.get_widget('name_entry')

        # System Detail Window Elements
        self.system_detail_title = self.wtree.get_widget('system_detail_title')
        self.sysdetail_coord_label = self.wtree.get_widget('sysdetail_coord_label')
        self.sysdetail_notes_heading = self.wtree.get_widget('sysdetail_notes_heading')
        self.sysdetail_notes_label = self.wtree.get_widget('sysdetail_notes_label')
        self.sysdetail_distance_heading = self.wtree.get_widget('sysdetail_distance_heading')
        self.sysdetail_distance_label = self.wtree.get_widget('sysdetail_distance_label')
        self.sysdetail_numplanet_label = self.wtree.get_widget('sysdetail_numplanet_label')
        self.sysdetail_mineral_value = self.wtree.get_widget('sysdetail_mineral_value')
        self.sysdetail_mineral_weight = self.wtree.get_widget('sysdetail_mineral_weight')
        self.sysdetail_mineral_worth = self.wtree.get_widget('sysdetail_mineral_worth')
        self.sysdetail_bio_value = self.wtree.get_widget('sysdetail_bio_value')
        self.sysdetail_spacing_hdr_label = self.wtree.get_widget('sysdetail_spacing_hdr_label')
        self.sysdetail_unfiltered_hdr_label = self.wtree.get_widget('sysdetail_unfiltered_hdr_label')
        self.sysdetail_filtered_hdr_label = self.wtree.get_widget('sysdetail_filtered_hdr_label')
        self.sysdetail_minbreakdown_blank_label = self.wtree.get_widget('sysdetail_minbreakdown_blank_label')
        self.sysdetail_minbreakdown_filt_label = self.wtree.get_widget('sysdetail_minbreakdown_filt_label')
        self.sysdetail_minbreakdown_unfilt_label = self.wtree.get_widget('sysdetail_minbreakdown_unfilt_label')
        self.sysdetail_filt_mineral_value = self.wtree.get_widget('sysdetail_filt_mineral_value')
        self.sysdetail_filt_mineral_weight = self.wtree.get_widget('sysdetail_filt_mineral_weight')
        self.sysdetail_filt_mineral_worth = self.wtree.get_widget('sysdetail_filt_mineral_worth')
        self.sysdetail_filt_bio_value = self.wtree.get_widget('sysdetail_filt_bio_value')
        self.sysdetail_safety_label = self.wtree.get_widget('sysdetail_safety_label')
        for min in min_vals:
            self.__dict__['sys_%s' % min] = self.wtree.get_widget('sys_%s' % min)
            self.__dict__['sys_filt_%s' % min] = self.wtree.get_widget('sys_filt_%s' % min)
        self.planet_detail = self.wtree.get_widget('planet_detail')

        # Set up columns for our planet details TreeView
        cols = ['Pos', 'Name', 'Type', 'Tectonics', 'Weather', 'Temp', 'Gravity', 'Bio', '(danger)', 'Value', 'Weight', 'MVpU']
        for (num, name) in enumerate(cols):
            renderer = gtk.CellRendererText()
            col = gtk.TreeViewColumn(name, renderer, text=num, foreground=20)
            col.set_sort_column_id(num)
            if (name == 'MVpU'):
                col.set_cell_data_func(renderer, cellformat_float, (num, '%0.1f'))
            self.planet_detail.append_column(col)
        for num, name in enumerate(min_vals):
            realnum = num + len(cols)
            col = gtk.TreeViewColumn(None, gtk.CellRendererText(), text=realnum, foreground=20)
            img = gtk.Image()
            img.set_from_file(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'min_%s.png' % (name)))
            img.set_tooltip_text(string.capitalize(name))
            img.show()
            col.set_widget(img)
            col.set_sort_column_id(realnum)
            self.planet_detail.append_column(col)

        # Column toggles
        self.planet_col_toggle_weather = self.wtree.get_widget('planet_col_toggle_weather')
        self.planet_col_toggle_bio = self.wtree.get_widget('planet_col_toggle_bio')
        self.planet_col_toggle_minsummary = self.wtree.get_widget('planet_col_toggle_minsummary')
        self.planet_col_toggle_mindetails = self.wtree.get_widget('planet_col_toggle_mindetails')

        # Populate planet type dropdown
        self.type_dropdown.remove_text(0)
        for type in sorted(planet_types.keys()):
            self.type_dropdown.append_text(type)

        # Populate constellation name dropdown
        self.const_dropdown.remove_text(0)
        for name in sorted(const_names.keys()):
            self.const_dropdown.append_text(name)

        # Set the initial state of our combo boxes
        self.safety_preset_box.set_active(0)
        self.tectonics_dir.set_active(0)
        self.weather_dir.set_active(0)
        self.temp_dir.set_active(0)
        self.bio_dir.set_active(0)
        self.type_dropdown.set_active(0)
        self.const_dropdown.set_active(0)

        # We're not currently setting a safety preset, or the name lookup
        self.setting_safety_preset = False
        self.setting_safety_custom = False
        self.setting_name_entry = False

        # Signals dictionary
        dic = { 'on_quit': self.on_quit,
                'on_about': self.on_about,
                'on_expose': self.on_expose,
                'on_window_resize': self.on_window_resize,
                'on_mouse_changed': self.on_mouse_changed,
                'on_map_clicked': self.on_map_clicked,
                'on_mouse_released': self.on_mouse_released,
                'on_trigger_redraw': self.on_trigger_redraw,
                'on_visvalue_radio_changed': self.on_visvalue_radio_changed,
                'on_dispspecial_checked': self.on_dispspecial_checked,
                'apply_filters': self.apply_filters,
                'on_proximity_val_changed': self.on_proximity_val_changed,
                'on_type_dropdown_changed': self.on_type_dropdown_changed,
                'on_const_dropdown_changed': self.on_const_dropdown_changed,
                'on_open_systemdetails_button': self.on_open_systemdetails_button,
                'on_planet_col_changed': self.on_planet_col_changed,
                'on_set_safety_preset': self.on_set_safety_preset,
                'on_safety_changed': self.on_safety_changed,
                'on_name_check_toggled': self.on_name_check_toggled,
                'on_name_entry_activate': self.on_name_entry_activate
                }
        self.wtree.signal_autoconnect(dic)

        # Vars for showing popups
        self.showing_popup = 0
        self.popup_x = -1
        self.popup_y = -1
        self.popup_width = -1
        self.popup_height = -1
        self.popup_spacing = 10
        self.popupsurf = None
        self.popupctx = None

        # Pango definitions which will never change
        self.pangoctx = self.window.get_pango_context()
        self.popup_hdr_layout = pango.Layout(self.pangoctx)
        self.popup_hdr_layout.set_font_description(pango.FontDescription('sans bold 11'))
        self.popup_hdr_layout.set_wrap(pango.WRAP_WORD_CHAR)
        self.popup_hdr_layout.set_width(350*pango.SCALE)
        self.popup_body_layout = pango.Layout(self.pangoctx)
        self.popup_body_layout.set_font_description(pango.FontDescription('sans normal 9'))
        self.popup_body_layout.set_wrap(pango.WRAP_WORD_CHAR)
        self.popup_body_layout.set_width(350*pango.SCALE)

        # Set default geometry hints
        # TODO: this could use testing
        self.window.set_geometry_hints(None, init_width, init_height, max_width, max_height)
        self.curx = -1
        self.cury = -1
        width = gtk.gdk.screen_width()
        height = gtk.gdk.screen_height()
        if (width > init_width):
            if (width > max_width):
                setwidth = max_width
            else:
                setwidth = width
        if (height > init_height):
            if (height > max_height):
                setheight = max_height
            else:
                setheight = height
        if (setwidth != init_width and setheight != init_height):
            self.window.resize(setwidth, setheight)

        # Vars relating to dragging
        self.dragging = False
        self.cursor_move_drag = gtk.gdk.Cursor(gtk.gdk.DOT)

        # Set up quasispace surfaces/contexts - we'll never need to regenerate
        # these, so there's no point in doing them inside on_trigger_redraw
        self.quasisurf = cairo.ImageSurface(cairo.FORMAT_ARGB32, quasi_dim, quasi_dim)
        self.quasictx = cairo.Context(self.quasisurf)
        self.quasictx.set_source_rgba(0, .4, 0, 1)
        self.quasictx.paint()
        self.quasictx.set_source_rgba(0, .6, 0, 1)
        for i in (range(1, 4)):
            self.quasictx.move_to(i*50+.5+quasi_line_offset, 0)
            self.quasictx.line_to(i*50+.5+quasi_line_offset, quasi_dim)
            self.quasictx.set_line_width(1)
            self.quasictx.stroke()
            self.quasictx.move_to(0, i*50+.5+quasi_line_offset)
            self.quasictx.line_to(200, i*50+.5+quasi_line_offset)
            self.quasictx.set_line_width(1)
            self.quasictx.stroke()
        self.quasictx.set_source_rgba(1, 1, 1, 1)
        self.quasictx.set_line_width(2)
        self.quasictx.rectangle(0, 0, quasi_dim, quasi_dim)
        self.quasictx.stroke()
        for quasi in self.systems.quasispace:
            self.quasictx.set_source_rgba(.01, .8, .01, 1)
            self.quasictx.arc(quasi.draw_quasi_x, quasi.draw_quasi_y, 2, 0, 2*math.pi)
            self.quasictx.fill()
        self.quasicachesurf = cairo.ImageSurface(cairo.FORMAT_ARGB32, quasi_dim, quasi_dim)
        self.quasicachectx = cairo.Context(self.quasicachesurf)
        self.quasicachectx.set_source_surface(self.quasisurf, 0, 0)
        self.quasicachectx.paint()

        # Set an idle func to init the map
        self.cursystem = -1
        self.curclicked = -1
        self.mapinit = False
        gobject.idle_add(self.on_trigger_redraw)

    def run(self):
        """
        Run the application
        """
        self.window.show()
        gtk.main()

    def on_trigger_redraw(self, widget=None):
        """
        Sets up the main map graphics.  There's a few surfaces/contexts flying around
        here, which could use some documentation.

        mapmap
        mapctx - The ctx directly tied to the main window.  All transient graphics (for
                 instance, mouseover'd systems and popups) are drawn directly to here,
                 and of course clicked systems get updated as appropriate (but are
                 drawn to activectx first).

        cachesurf
        cachectx - This is a cache, and the surface where we initially build the map.
                   It is *never* updated once the initial map has been constructed, and
                   is used as a "clean" surface to grab image data from (for instance, when
                   the mouse moves away from a system, so we need to un-highlight it).

        activesurf
        activectx - This is a copy of the basic map but with any semi-permanent embellishments
                    added in.  For instance, the currently-clicked on star.  This is the
                    surface which will be used to update the main screen when transient graphics
                    (mouseovers, popups) go away.  At the moment it's sort of overkill because
                    the only difference between this and cachesurf is the currently-clicked-on
                    system, but in the future there may be more, so we'll leave it.
        """

        # Vars
        self.mainwin = self.mainarea.window
        self.gc_state = self.mainarea.get_style().fg_gc[gtk.STATE_NORMAL]
        self.mapmap = gtk.gdk.Pixmap(self.mainarea.window, 1001, 1001, -1)
        self.mapctx = self.mapmap.cairo_create()

        self.cachesurf = cairo.ImageSurface(cairo.FORMAT_ARGB32, 1001, 1001)
        self.cachectx = cairo.Context(self.cachesurf)

        # Colors
        self.c_background = (0, 0, 0, 1)
        self.c_grid = (.4, .4, .4, 1)
        self.c_histar = (1, 1, 1, 1)
        self.c_regstar = (.5, .5, .5, 1)
        self.c_hispecstar = (1, .4, .4, 1)
        self.c_specstar = (.6, .4, .4, 1)
        self.c_curstar = (.4, 1, .4, 1)
        self.c_curstar_vis = self.c_histar
        self.c_outline = (.5, .5, .6, 1)
        self.c_quasi = (0, 1, 0, 1)
        self.c_const = (.5, .5, .5, 1)

        # Draw the map lines
        self.cachectx.set_source_rgba(0, 0, 0, 1)
        self.cachectx.paint()
        for i in (range(1,10)):
            self.cachectx.set_source_rgba(*self.c_grid)
            self.cachectx.move_to(i*100+.5, 0)
            self.cachectx.line_to(i*100+.5, 1001)
            self.cachectx.set_line_width(1)
            self.cachectx.stroke()
            self.cachectx.move_to(0, i*100+.5)
            self.cachectx.line_to(1001, i*100+.5)
            self.cachectx.set_line_width(1)
            self.cachectx.stroke()

        # Draw our constellation connections, if we're doing that
        if self.dispconst_check.get_active():
            for (sys_from, sys_to) in self.systems.connections:
                self.cachectx.set_source_rgba(*self.c_const)
                self.cachectx.move_to(sys_from[0], sys_from[1])
                self.cachectx.line_to(sys_to[0], sys_to[1])
                self.cachectx.set_line_width(1)
                self.cachectx.stroke()

        # Render a circle denoting our proximity, if we're doing that
        if self.proximity_check.get_active():
            system = self.systems.get(self.curclicked)
            if system:
                self.cachectx.set_source_rgba(*self.c_outline)
                self.cachectx.arc(system.draw_x, system.draw_y, self.proximity_spin.get_value_as_int(), 0, 2*math.pi)
                self.cachectx.set_line_width(1)
                self.cachectx.stroke()

        # Render systems
        visvalue = self.visvalue_check.get_active()
        for system in self.systems.getall():
            self.draw_star(system, True)

        # Grab a backup of the cachectx which we'll use for our currently-clicked star data
        # This is a bit overkill; all we *really* need is a small surface enough to hold some
        # backup data and we'd be good...
        self.activesurf = cairo.ImageSurface(cairo.FORMAT_ARGB32, 1001, 1001)
        self.activectx = cairo.Context(self.activesurf)
        self.activectx.set_source_surface(self.cachesurf, 0, 0)
        self.activectx.paint()

        # Render and let the rest of the prog know we're initted
        self.mapctx.set_source_surface(self.cachesurf, 0, 0)
        self.mapctx.paint()
        self.mainwin.draw_drawable(self.gc_state, self.mapmap, 0, 0, 0, 0, 1001, 1001)
        self.mapinit = True

        # We call this so that our currently-clicked star remains looking currently-clicked
        # TODO: there's a "flash" here as this updates; figure that out.
        self.process_clicked_gfx()

        # so we only run once when called from idle_add
        return False

    def on_quit(self, widget=None, event=None):
        """
        Quit the application
        """
        gtk.main_quit()

    def on_about(self, widget):
        """
        Throw up an About screen
        """
        if (not self.aboutwindow):
            about = gtk.AboutDialog()
            about.set_transient_for(self.window)
            about.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
            about.set_name(app_name)
            about.set_version(app_version)
            about.set_website(app_url)
            about.set_authors(['Code - CJ Kucera', 'Mineral Icons, Map Data - Toys For Bob / Ur-Quan Masters team'])
            licensepath = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'COPYING.txt')
            if (os.path.isfile(licensepath)):
                try:
                    df = open(licensepath, 'r')
                    about.set_license(df.read())
                    df.close()
                except:
                    pass
            self.aboutwindow = about

        # Now run
        self.aboutwindow.run()
        self.aboutwindow.hide()

    def on_expose(self, widget, event):
        """
        What to do when our map is exposed (ie: just fill in data from our surface)
        """
        if (self.mapinit):
            self.mainwin.draw_drawable(self.gc_state, self.mapmap, 0, 0, 0, 0, 1001, 1001)

    def on_window_resize(self, widget, rect):
        """
        I want the window to not grow beyond the total width/height of the map.
        Unfortunately if we let gtk itself take care of the scrollbars, they'll end
        up staying present at full size.  So, we'll just check in here to see if we've
        been resized to the full width/height, and if so disable the appropriate
        scrollbars.  Alas!  Note that because we update the coordinate label with our
        current cursor position in on_mouse_changed, this will get called with every
        change of the mouse, so we keep track of our current dimensions and return
        right away if they haven't actually changed.
        """
        if (self.curx == rect.width and self.cury == rect.height):
            return
        if (rect.width == max_width):
            hpolicy = gtk.POLICY_NEVER
        else:
            hpolicy = gtk.POLICY_AUTOMATIC
        if (rect.height == max_height):
            vpolicy = gtk.POLICY_NEVER
        else:
            vpolicy = gtk.POLICY_AUTOMATIC
        self.mainscroll.set_policy(hpolicy, vpolicy)
        self.curx = rect.width
        self.cury = rect.height

    def update_star(self, x, y, fromsurf, toctx=None):
        """
        This is used to update star graphics when various things happen
        (like mouseovers or clicks).  It updates the appropriate region
        from "fromsurf" and copies it to "toctx".
        """
        if not toctx:
            toctx = self.mapctx
        startx = int(x/10)-5
        starty = 1000-int(y/10)-5
        width = 10
        height = 10
        if (startx < 0):
            width = width+startx
            startx = 0
        if (starty < 0):
            height = height+starty
            starty = 0
        toctx.save()
        toctx.set_operator(cairo.OPERATOR_SOURCE)
        toctx.rectangle(startx, starty, width, height)
        toctx.set_source_surface(fromsurf, 0, 0)
        toctx.fill()
        toctx.restore()
        if (toctx == self.mapctx):
            self.mainarea.queue_draw()

    def clean_star(self, id):
        """
        Cleans a star (typically after the mouse has moved away) by copying over
        data from activesurf.  We use that instead of cachesurf because that way
        we're filling in the currently-clicked image properly
        """
        system = self.systems.get(id)
        if system:
            self.update_star(system.x, system.y, self.activesurf)

    def draw_star(self, system, initial=False):
        """
        Draws a single star.  When run during the initial population, this will just draw the
        basic starmap, when run afterwards it'll draw the highlighted version.
        """
        if (system.quasi):
            if self.quasi_check.get_active():
                doquasi = True
            else:
                return
        else:
            doquasi = False
        if system:
            if initial:
                if (doquasi):
                    rad = 3
                else:
                    rad = 2
                ctx = self.cachectx
            else:
                rad = 4
                ctx = self.mapctx

            if (not initial and self.curclicked == system.id):
                color = self.c_curstar
            elif (doquasi):
                color = self.c_quasi
                ctx.set_line_width(2)
            elif (self.special_check.get_active() and system.extra and system.extra != ''):
                if (system.highlight):
                    color = self.c_hispecstar
                    if (self.visvalue_check.get_active()):
                        if initial:
                            rad = 4
                        else:
                            rad = 5
                else:
                    color = self.c_specstar
            elif (system.highlight):
                color = self.c_histar
                if (self.visvalue_check.get_active()):
                    if (self.visvalue_mineral_radio.get_active()):
                        intens = self.systems.ret_intensity(system)
                    else:
                        intens = self.systems.ret_bio_intensity(system)
                    # .6 and above is quite sparse, should just be hot
                    # .4 to .6 is even quite sparse
                    # .35 and below is where the vast majority of the systems lie
                    if (intens > .8):
                        color = (1, 1, 1, 1)
                    elif (intens > .6):
                        #intens = (intens-.6)/.4
                        #color = colorsys.hsv_to_rgb(0, 1-intens, 1)
                        color = (1, 0, 0, 1)
                    else:
                        intens = intens/.6
                        # 160 seems all right on a linear scale; really it should be sort-of parametric or something
                        color = colorsys.hsv_to_rgb((160-(160*intens))/256.0, 1, 1)
            else:
                color = self.c_regstar

            # Do the actual drawing
            ctx.set_source_rgba(*color)
            ctx.arc(system.draw_x, system.draw_y, rad, 0, 2*math.pi)
            if (doquasi):
                ctx.stroke()
            else:
                ctx.fill()

    def process_clicked_gfx(self, system=None):
        """
        Function to deal with upgrading the "current" surface with the highlighted star.
        Technically we're overlapping a bit with draw_star, but not by much.
        """
        if not system:
            system = self.systems.get(self.curclicked)
        if system:
            if (self.visvalue_check.get_active()):
                rad = 4
                color = self.c_curstar_vis
            else:
                rad = 2
                color = self.c_curstar
            self.activectx.set_source_rgba(*color)
            self.activectx.arc(system.draw_x, system.draw_y, rad, 0, 2*math.pi)
            self.activectx.fill()
            self.update_star(system.x, system.y, self.activesurf)

    def render_popup(self):
        """
        Renders the actual popup to mapctx.  This assumes that all the relevant
        popup_* vars are set and accurate.
        """
        if self.showing_popup:
            self.mapctx.save()
            self.mapctx.set_operator(cairo.OPERATOR_SOURCE)
            self.mapctx.rectangle(self.popup_x, self.popup_y, self.popup_width, self.popup_height)
            self.mapctx.set_source_surface(self.popupsurf, self.popup_x, self.popup_y)
            self.mapctx.fill()
            self.mapctx.restore()

    def launch_popup(self, system):
        """
        Launch a popup with basic information about the given system
        """
        if (not self.popup_check.get_active()):
            return
        self.discard_popup()
        sys_x = system.draw_x
        sys_y = system.draw_y

        # Set up our header text
        self.popup_hdr_layout.set_markup('<b>%s</b>' % (system.fullname))
        (hdr_width, hdr_height) = map(lambda x: x/pango.SCALE, self.popup_hdr_layout.get_size())

        # Set up the body text
        bodytext = []
        bodytext.append('<b>Position:</b> (%0.1f, %0.1f)' % (system.x/10.0, system.y/10.0))
        if not system.quasi:
            if (self.dispspecial_check.get_active() and system.extra and system.extra != ''):
                bodytext.append('<b>Notes:</b> %s' % (system.extra))
            bodytext.append('<b>Planets:</b> %d' % (len(system.planets)))
            if (self.systems.aggfilter.filtering()):
                if (system.agg_value() != system.full_agg_value()):
                    bodytext.append('<b>Mineral Value:</b> %d  <small>(%d unfiltered)</small>' % (system.agg_value(), system.full_agg_value()))
                else:
                    bodytext.append('<b>Mineral Value:</b> %d  <small>(unchanged by filter)</small>' % (system.agg_value()))
                if (system.agg_worth() != system.full_agg_worth()):
                    bodytext.append('<b>Min. Value per Unit:</b> %0.1f  <small>(%0.1f unfiltered)</small>' % (system.agg_worth(), system.full_agg_worth()))
                else:
                    bodytext.append('<b>Min. Value per Unit:</b> %0.1f  <small>(unchanged by filter)</small>' % (system.agg_worth()))
                if (system.bio_agg_value() != system.full_bio_agg_value()):
                    bodytext.append('<b>Bio Value:</b> %d  <small>(%d unfiltered)</small>' % (system.bio_agg_value(), system.full_bio_agg_value()))
                else:
                    bodytext.append('<b>Bio Value:</b> %d  <small>(unchanged by filter)</small>' % (system.bio_agg_value()))
            else:
                bodytext.append('<b>Mineral Value:</b> %d' % (system.agg_value()))
                bodytext.append('<b>Min. Value per Unit:</b> %0.1f' % (system.agg_worth()))
                if (system.bio_danger_agg_value() == 0):
                    bodytext.append('<b>Bio Value:</b> %d' % (system.bio_agg_value()))
                else:
                    bodytext.append('<b>Bio Value:</b> %d  <small>(%d dangerous)</small>' % (system.bio_agg_value(), system.bio_danger_agg_value()))
        if (self.curclicked and self.curclicked != -1 and system.id != self.curclicked):
            origin = self.systems.get(self.curclicked)
            bodytext.append('<b>Distance from Selected:</b> %0.1f' % (system.distance_to(origin)))
        if not system.quasi:
            bodytext.append("<small><small>\n(Right-click or Double-click for details)</small></small>")
        self.popup_body_layout.set_markup("\n".join(bodytext))
        (body_width, body_height) = map(lambda x: x/pango.SCALE, self.popup_body_layout.get_size())

        # Figure out our actual dimensions
        if (hdr_width > body_width):
            text_width = hdr_width
        else:
            text_width = body_width

        content_width = text_width
        self.popup_width = text_width+(self.popup_spacing*2)
        self.popup_height = hdr_height+body_height+(self.popup_spacing*2)+11

        # ... adjust for if we're quasispace or not
        if system.quasi:
            self.popup_height += quasi_dim+(self.popup_spacing*2)
            if (self.popup_width < (quasi_dim + self.popup_spacing*2)):
                self.popup_width = quasi_dim + (self.popup_spacing*2)
                content_width = quasi_dim
            quasi_start_x = self.popup_spacing + ((content_width-quasi_dim)/2)
            quasi_start_y = (self.popup_spacing*2) + body_height + hdr_height + 11

        # Figure out the X placement of the popup
        adjust = self.mainscroll.get_hadjustment()
        if (sys_x - adjust.value + (self.popup_spacing * 2) + self.popup_width > adjust.page_size):
            self.popup_x = sys_x - self.popup_spacing - self.popup_width
        else:
            self.popup_x = sys_x + self.popup_spacing

        # Figure out the Y placement of the popup
        adjust = self.mainscroll.get_vadjustment()
        if (sys_y - adjust.value + (self.popup_spacing * 2) + self.popup_height > adjust.page_size):
            self.popup_y = sys_y - self.popup_spacing - self.popup_height
        else:
            self.popup_y = sys_y + self.popup_spacing

        # Let's open up a context to work with
        self.popupsurf = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.popup_width, self.popup_height)
        self.popupctx = cairo.Context(self.popupsurf)
        
        # Initialize
        self.popupctx.set_source_rgba(.1, .1, .1, 1)
        self.popupctx.paint()

        # Draw a border
        self.popupctx.set_source_rgba(1, 1, 1, 1)
        self.popupctx.rectangle(0, 0, self.popup_width, self.popup_height)
        self.popupctx.set_line_width(1)
        self.popupctx.stroke()

        # Now render our header text
        self.popupctx.move_to(self.popup_spacing+((content_width-hdr_width)/2), self.popup_spacing)
        pangoctx = pangocairo.CairoContext(self.popupctx)
        pangoctx.show_layout(self.popup_hdr_layout)

        # ... and now the body
        self.popupctx.move_to(self.popup_spacing, self.popup_spacing + hdr_height + 11)
        pangoctx.show_layout(self.popup_body_layout)

        # Draw a line inbetween.  Gosh we're fancy.
        self.popupctx.move_to(self.popup_spacing, self.popup_spacing + hdr_height + 5)
        self.popupctx.line_to(self.popup_width - self.popup_spacing, self.popup_spacing + hdr_height + 5)
        self.popupctx.set_source_rgba(1, 1, 1, 1)
        self.popupctx.set_line_width(.5)
        self.popupctx.stroke()

        # ... if we're quasispace, draw the map
        if system.quasi:
            # First refresh with the cached map
            self.quasictx.set_source_surface(self.quasicachesurf, 0, 0)
            self.quasictx.paint()

            # Now highlight the proper exit
            self.quasictx.set_source_rgba(1, 1, 1, 1)
            self.quasictx.arc(system.draw_quasi_x, system.draw_quasi_y, 3, 0, 2*math.pi)
            self.quasictx.fill()

            # ... and now render into the popup
            self.popupctx.save()
            self.popupctx.set_operator(cairo.OPERATOR_SOURCE)
            self.popupctx.rectangle(quasi_start_x, quasi_start_y, quasi_dim, quasi_dim)
            self.popupctx.set_source_surface(self.quasisurf, quasi_start_x, quasi_start_y)
            self.popupctx.fill()
            self.popupctx.restore()

        # Flip our boolean so that we know that we're showing the popup
        self.showing_popup = True

        # copy it over to our context
        self.render_popup()

    def discard_popup(self):
        """
        Discards the currently-active popup
        """
        if (self.showing_popup):
            toctx = self.mapctx
            toctx.save()
            toctx.set_operator(cairo.OPERATOR_SOURCE)
            toctx.rectangle(self.popup_x, self.popup_y, self.popup_width, self.popup_height)
            toctx.set_source_surface(self.activesurf, 0, 0)
            toctx.fill()
            toctx.restore()
            self.mainarea.queue_draw()
            self.showing_popup = False

    def on_mouse_changed(self, widget, event):
        """
        Track the mouse as it moves around the starmap, or drag the starmap
        if that's what we're up to.
        """
        if (self.dragging):
            diff_x = self.hold_x - event.x_root
            diff_y = self.hold_y - event.y_root
            if (diff_x != 0):
                adjust = self.mainscroll.get_hadjustment()
                newvalue = adjust.get_value() + diff_x
                if (newvalue < adjust.lower):
                    newvalue = adjust.lower
                elif (newvalue > adjust.upper-adjust.page_size):
                    newvalue = adjust.upper-adjust.page_size
                adjust.set_value(newvalue)
            if (diff_y != 0):
                adjust = self.mainscroll.get_vadjustment()
                newvalue = adjust.get_value() + diff_y
                if (newvalue < adjust.lower):
                    newvalue = adjust.lower
                elif (newvalue > adjust.upper-adjust.page_size):
                    newvalue = adjust.upper-adjust.page_size
                adjust.set_value(newvalue)
            self.hold_x = event.x_root
            self.hold_y = event.y_root
            return

        # Theoretically we don't need these bounds checks since we're now limiting
        # window dimensions.  Still, better safe than sorry, yes?
        if (event.x > 1000):
            event_x = 1000
        elif (event.x < 0):
            event_x = 0
        else:
            event_x = event.x
        if (event.y > 1000):
            event_y = 1000
        elif (event.y < 0):
            event_y = 0
        else:
            event_y = event.y
        x = event_x
        y = 1000-event_y
        for i in range(5):
            for j in range(5):
                # A fun little bug!  Due to the way we do these lookups, it's
                # possible to select Zeta Serpentis (500.7, 3.5) from the other
                # side of the map, way up at the top.  It's due to Python's
                # negative-array-index handling.  I don't really care enough to
                # put in proecssing to fix it, given that it only affects one
                # system.  Just enjoy the wormhole!  :)
                system = self.systems.lookup(event_x-i, event_y-j)
                if system:
                    if (system.quasi and not self.quasi_check.get_active()):
                        continue
                    if (self.cursystem != system.id):
                        if (self.cursystem != -1):
                            self.clean_star(self.cursystem)
                        self.draw_star(system)
                        self.cursystem = system.id
                        self.infolabel.set_markup('<i>%s - (%d, %d)</i>' % (system.fullname, x, y))
                        self.launch_popup(system)
                        self.mainarea.queue_draw()
                    return
        if (self.cursystem != -1):
            self.clean_star(self.cursystem)
            self.cursystem = -1
            self.discard_popup()
        self.infolabel.set_markup('<i>(%d, %d)</i>' % (x, y))

    def on_map_clicked(self, widget, event):
        """
        Handle a click, used to select a system
        Also a double-click or right-click to bring up info
        """
        if (event.type == gtk.gdk._2BUTTON_PRESS or event.button == 3):
            self.open_systemdetails(self.cursystem)
        elif (event.button == 1 and self.cursystem == -1 and (self.curx != max_width or self.cury != max_height)):
            self.dragging = True
            self.hold_x = event.x_root
            self.hold_y = event.y_root
            self.diff_x = 0
            self.diff_y = 0
            self.mainarea.window.set_cursor(self.cursor_move_drag)
        elif (self.cursystem != -1):
            system = self.systems.get(self.cursystem)
            if (self.curclicked != self.cursystem and self.curclicked != -1):
                oldsystem = self.systems.get(self.curclicked)
                if oldsystem:
                    self.update_star(oldsystem.x, oldsystem.y, self.cachesurf, self.activectx)
                    self.clean_star(oldsystem.id)
            self.curclicked = self.cursystem

            # Update graphics
            # Note that we re-display our popup, if we have one, so the previously-clicked-on
            # star doesn't display "over" the popup, if it was previously under
            self.process_clicked_gfx(system)
            self.render_popup()
            self.draw_star(system)

            # Update labels
            self.update_selected_labels(system)

            # Now make sure we redraw stuff as-necessary, which should be only when we have
            # the proximity filter on.
            if (self.proximity_check.get_active()):
                self.apply_filters()

    def on_mouse_released(self, widget, event):
        """
        What do to when the mouse button is lifted up.  No effect unless
        we're currently dragging the map.
        """
        if (self.dragging):
            self.dragging = False
            self.mainarea.window.set_cursor(None)

    def update_selected_labels(self, system):
        """
        Updates the text labels on the main screen which describe the currently-clicked-on system
        """
        self.namelabel.set_markup('<small><i>%s</i></small>'  % system.fullname)
        self.coordlabel.set_markup("<small>(%0.1f, %0.1f)</small>" % (system.x/10.0, system.y/10.0))
        self.coordlabel.show()
        if (system.quasi):
            self.systemdetails_button.hide()
        else:
            self.systemdetails_button.show()

    def on_visvalue_radio_changed(self, widget=None):
        """
        What to do when we've changed our visual value radio button.  Instead of
        blindly calling an on_trigger_redraw, we'll make sure that we're even using
        that option first.
        """
        if self.visvalue_check.get_active():
            self.on_trigger_redraw()

    def apply_filters(self, widget=None):
        """
        Applies whatever viewing filters are active.  First strips all highlighting.
        """

        # Reset our filters
        self.systems.aggfilter.reset()
        self.systems.dispfilter.reset()

        # First add a proximity filter if we need to
        if (self.proximity_check.get_active()):
            self.systems.dispfilter.add(ProxFilter(self.systems.get(self.curclicked), self.proximity_spin.get_value_as_int()))

        # Next set up a filter based on planet type
        if (self.type_check.get_active()):
            self.systems.dispfilter.add(TypeFilter(self.type_dropdown.get_active_text()))

        # Constellation name
        if (self.const_check.get_active()):
            self.systems.dispfilter.add(ConstFilter(self.const_dropdown.get_active_text()))

        # Now one based on system name
        if (self.name_check.get_active() and self.name_entry.get_text() != ''):
            self.systems.dispfilter.add(NameFilter(self.name_entry.get_text(), self.dispspecial_check.get_active()))

        # Now process the safety filters.
        if (self.safety_check.get_active()):
            # We shouldn't have to do ALL of this every time, but it really
            # shouldn't take any resources to speak of, so we'll just be thorough about it
            safetyfilter = SafetyFilter()
            if (self.tectonics_dir.get_active() == 1):
                safetyfilter.tectonics_dir = 1
            if (self.weather_dir.get_active() == 1):
                safetyfilter.weather_dir = 1
            if (self.temp_dir.get_active() == 1):
                safetyfilter.temp_dir = 1
            if (self.bio_dir.get_active() == 1):
                safetyfilter.bio_dir = 1
            safetyfilter.tectonics_val = self.tectonics_spin.get_value_as_int()
            safetyfilter.weather_val = self.weather_spin.get_value_as_int()
            safetyfilter.temp_val = self.temp_spin.get_value_as_int()
            safetyfilter.bio_val = self.bio_spin.get_value_as_int()
            safetyfilter.activate()
            self.systems.aggfilter.add(safetyfilter)

        # For completeness' sake, let's just *always* process aggregates here...
        self.systems.process_aggregates()

        # ... and now update our graphics
        self.on_trigger_redraw()

    def on_proximity_val_changed(self, widget=None):
        """
        The user changed the proximity spin button
        """
        if (self.proximity_check.get_active()):
            self.apply_filters()

    def on_type_dropdown_changed(self, widget=None):
        """
        The user selected a new planet type to (possibly) filter on
        """
        if (self.type_check.get_active()):
            self.apply_filters()

    def on_const_dropdown_changed(self, widget=None):
        """
        The user selected a new constellation name to (possibly) filter on
        """
        if (self.const_check.get_active()):
            self.apply_filters()

    def on_name_check_toggled(self, widget=None):
        """
        The user checked the plaintext name search filter
        """
        if self.setting_name_entry:
            return
        if (self.name_entry.get_text() != ''):
            self.apply_filters()

    def on_name_entry_activate(self, widget):
        """
        This fires when a user hits "enter" inside the system name text area.
        We'll automatically toggle the checkbox if appropriate, or unselect it.
        """
        if (widget.get_text() != '' and not self.name_check.get_active()):
            self.setting_name_entry = True
            self.name_check.set_active(True)
            self.setting_name_entry = False
        elif (widget.get_text() == '' and self.name_check.get_active()):
            self.setting_name_entry = True
            self.name_check.set_active(False)
            self.setting_name_entry = False
        elif not self.name_check.get_active():
            return
        self.apply_filters()

    def on_dispspecial_checked(self, widget=None):
        """
        What to do when we toggle the "display special" checkmark.  For the most part
        this doesn't have immediate effects; popups and the data screen will read the
        value as-needed to find out what to draw.  However, if we're searching based
        on system name, we'll have to update our results list, since that filter will
        search on the 'extra' attribute if we're checked.  That way searches for, say,
        'Melnorme', will complete properly.
        """
        if (self.name_check.get_active()):
            self.apply_filters()

    def on_set_safety_preset(self, widget):
        """
        Sets the various safety values to a particular preset.
        The code is pretty particular about the order of this combobox.
        The elements should be:

          0 - Completely-safe environments
          1 - Reasonably-safe environments
          2 - Lifeform resistant
          3 - Earthquake-resistant lander
          4 - Lightning-resistant lander
          5 - Fire-resistant lander
          6 - All
          7 - Custom (only present when the user has been fiddling)

        The various levels are intended to sort of generally follow the
        path of lander upgrades that you'll get via the Melnorme, though
        I have yet to test this out at various stages of the game to see
        how resasonable they are.

        Speaking of "Reasonable," the distinction between 0 and 1 is that
        the completely-safe preset will feature absolutely no threats
        whatsoever, whereas the "reasonably-safe" preset will have some
        earthquakes and fires and the like, but at levels that I think
        are generally avoidable if you remain cagey.
        """
        # TODO: Bio danger should be included in all of these
        if self.setting_safety_custom:
            return
        if (len(self.safety_preset_box.get_model()) == max_presets + 1):
            self.safety_preset_box.remove_text(max_presets)
        self.setting_safety_preset = True
        self.tectonics_dir.set_active(0)
        self.weather_dir.set_active(0)
        self.temp_dir.set_active(0)
        self.bio_dir.set_active(0)
        if (widget.get_active() == 0):
            # Completely-safe
            self.tectonics_spin.set_value(2)
            self.weather_spin.set_value(2)
            self.temp_spin.set_value(100)
            self.bio_spin.set_value(0)
        elif (widget.get_active() == 1):
            # Reasonably safe
            self.tectonics_spin.set_value(5)
            self.weather_spin.set_value(2)
            self.temp_spin.set_value(400)
            self.bio_spin.set_value(10)
        elif (widget.get_active() == 2):
            # Lifeform Resistant
            self.tectonics_spin.set_value(5)
            self.weather_spin.set_value(2)
            self.temp_spin.set_value(400)
            self.bio_spin.set_value(300)
        elif (widget.get_active() == 3):
            # Earthquake-Resistant
            self.tectonics_spin.set_value(7)
            self.weather_spin.set_value(2)
            self.temp_spin.set_value(400)
            self.bio_spin.set_value(300)
        elif (widget.get_active() == 4):
            # Lightning-Resistant
            self.tectonics_spin.set_value(7)
            self.weather_spin.set_value(7)
            self.temp_spin.set_value(400)
            self.bio_spin.set_value(300)
        elif (widget.get_active() == 5):
            # Fire-Resistant
            self.tectonics_spin.set_value(7)
            self.weather_spin.set_value(7)
            self.temp_spin.set_value(2000)
            self.bio_spin.set_value(300)
        elif (widget.get_active() == 6):
            # All
            self.tectonics_spin.set_value(8)
            self.weather_spin.set_value(8)
            self.temp_spin.set_value(5200)
            self.bio_spin.set_value(300)
        self.setting_safety_preset = False
        if (self.safety_check.get_active()):
            self.apply_filters()

    def on_safety_changed(self, widget=None):
        """
        Process a change to a specific safety filter.  Note that this
        will flip the preset dropdown to "custom"
        """
        if (self.setting_safety_preset):
            return
        if (len(self.safety_preset_box.get_model()) == max_presets):
            self.setting_safety_custom = True
            self.safety_preset_box.append_text('Custom')
            self.safety_preset_box.set_active(max_presets)
            self.setting_safety_custom = False
        if (self.safety_check.get_active()):
            self.apply_filters()

    def on_open_systemdetails_button(self, widget=None):
        """
        The user clicked the small button in the upper-left, to get system details.
        """
        self.open_systemdetails(self.curclicked)

    def open_systemdetails(self, id):
        """
        Opens the main data window for a system, given the system's ID.
        """
        system = self.systems.get(id)
        if system and not system.quasi:
            sysname = system.fullname
            self.systemwindow.set_title('System Details: %s' % (sysname))
            self.system_detail_title.set_markup('<b>System Details: <i>%s</i></b>' % (sysname))
            self.sysdetail_coord_label.set_text('(%0.1f, %0.1f)' % (system.x/10.0, system.y/10.0))
            self.sysdetail_numplanet_label.set_text('%d' % (len(system.planets)))
            if (self.dispspecial_check.get_active() and system.extra and system.extra != ''):
                self.sysdetail_notes_heading.show()
                self.sysdetail_notes_label.show()
                self.sysdetail_notes_label.set_text(system.extra)
            else:
                self.sysdetail_notes_heading.hide()
                self.sysdetail_notes_label.hide()
            if (id != self.curclicked):
                clickedsys = self.systems.get(self.curclicked)
                if (clickedsys):
                    self.sysdetail_distance_heading.show()
                    self.sysdetail_distance_label.show()
                    self.sysdetail_distance_label.set_text('%0.1f units' % (system.distance_to(clickedsys)))
                else:
                    self.sysdetail_distance_heading.hide()
                    self.sysdetail_distance_label.hide()
            else:
                self.sysdetail_distance_heading.hide()
                self.sysdetail_distance_label.hide()

            # Now the "full" mineral/bio values
            self.sysdetail_mineral_value.set_text('%d' % system.full_agg_value())
            self.sysdetail_mineral_weight.set_text('%d' % system.full_agg_weight())
            self.sysdetail_mineral_worth.set_text('%0.1f' % system.full_agg_worth())
            if (system.full_bio_danger_agg_value() == 0):
                self.sysdetail_bio_value.set_text('%d' % system.full_bio_agg_value())
            else:
                self.sysdetail_bio_value.set_text('%d (%d dangerous)' % (system.full_bio_agg_value(), system.full_bio_danger_agg_value()))
            for min in min_vals:
                self.__dict__['sys_%s' % min].set_text('%d' % system.min_agg_full.__dict__[min])

            # Now, if required, deal with our filtered values
            if self.systems.aggfilter.filtering():
                self.sysdetail_filt_mineral_value.set_text('%d' % system.agg_value())
                self.sysdetail_filt_mineral_weight.set_text('%d' % system.agg_weight())
                self.sysdetail_filt_mineral_worth.set_text('%0.1f' % system.agg_worth())
                if (system.bio_danger_agg_value() == 0):
                    self.sysdetail_filt_bio_value.set_text('%d' % system.bio_agg_value())
                else:
                    self.sysdetail_filt_bio_value.set_text('%d (%d dangerous)' % (system.bio_agg_value(), system.bio_danger_agg_value()))
                self.sysdetail_spacing_hdr_label.show()
                self.sysdetail_unfiltered_hdr_label.show()
                self.sysdetail_filtered_hdr_label.show()
                self.sysdetail_minbreakdown_blank_label.show()
                self.sysdetail_minbreakdown_filt_label.show()
                self.sysdetail_minbreakdown_unfilt_label.show()
                self.sysdetail_safety_label.show()
                self.sysdetail_filt_mineral_value.show()
                self.sysdetail_filt_mineral_weight.show()
                self.sysdetail_filt_mineral_worth.show()
                self.sysdetail_filt_bio_value.show()
                self.sysdetail_safety_label.show()
                pairs = [ (self.sysdetail_bio_value, self.sysdetail_filt_bio_value),
                        (self.sysdetail_mineral_value, self.sysdetail_filt_mineral_value),
                        (self.sysdetail_mineral_weight, self.sysdetail_filt_mineral_weight),
                        (self.sysdetail_mineral_worth, self.sysdetail_filt_mineral_worth) ]
                for min in min_vals:
                    self.__dict__['sys_filt_%s' % min].show()
                    self.__dict__['sys_filt_%s' % min].set_text('%d' % system.min_agg.__dict__[min])
                    pairs.append((self.__dict__['sys_%s' % min], self.__dict__['sys_filt_%s' % min]))
                for pair in pairs:
                    if (pair[0].get_label() != pair[1].get_label()):
                        pair[0].set_markup('<span foreground="red">%s</span>' % pair[0].get_label())
            else:
                self.sysdetail_spacing_hdr_label.hide()
                self.sysdetail_unfiltered_hdr_label.hide()
                self.sysdetail_filtered_hdr_label.hide()
                self.sysdetail_minbreakdown_blank_label.hide()
                self.sysdetail_minbreakdown_filt_label.hide()
                self.sysdetail_minbreakdown_unfilt_label.hide()
                self.sysdetail_safety_label.hide()
                self.sysdetail_filt_mineral_value.hide()
                self.sysdetail_filt_mineral_weight.hide()
                self.sysdetail_filt_mineral_worth.hide()
                self.sysdetail_filt_bio_value.hide()
                self.sysdetail_safety_label.hide()
                for min in min_vals:
                    self.__dict__['sys_filt_%s' % min].hide()

            store = gtk.ListStore(
                    gobject.TYPE_INT,
                    gobject.TYPE_STRING,
                    gobject.TYPE_STRING,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_FLOAT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    gobject.TYPE_INT,
                    str)
            for num, planet in enumerate(system.planets):
                iter = store.append()
                #if (self.aggfilter.filtering() and not self.aggfilter.approve(planet)):
                if (not self.systems.aggfilter.approve(planet)):
                    fgcolor = 'red'
                else:
                    fgcolor = None
                store.set(iter,
                        0, num+1,
                        1, planet.name,
                        2, planet.type,
                        3, planet.tectonics,
                        4, planet.weather,
                        5, planet.temp,
                        6, planet.gravity,
                        7, planet.bio,
                        8, planet.bio_danger,
                        9, planet.mineral.value(),
                        10, planet.mineral.weight(),
                        11, planet.mineral.worth(),
                        12, planet.mineral.common,
                        13, planet.mineral.corrosive,
                        14, planet.mineral.base,
                        15, planet.mineral.noble,
                        16, planet.mineral.rare,
                        17, planet.mineral.precious,
                        18, planet.mineral.radioactive,
                        19, planet.mineral.exotic,
                        20, fgcolor)
            self.planet_detail.set_model(store)

            self.on_planet_col_changed()

            self.systemwindow.run()
            self.systemwindow.hide()

    def on_planet_col_changed(self, widget=None):
        """
        This is fired off when the user clicks any of the toggle buttons in
        the system detail window (ie: "biological data," "mineral details,"
        etc.  It basically just controls which columns are active.
        """
        weather = self.planet_col_toggle_weather.get_active()
        bio = self.planet_col_toggle_bio.get_active()
        minsummary = self.planet_col_toggle_minsummary.get_active()
        mindetails = self.planet_col_toggle_mindetails.get_active()
        for i in range(4):
            self.planet_detail.get_column(3+i).set_visible(weather)
        for i in range(2):
            self.planet_detail.get_column(7+i).set_visible(bio)
        for i in range(3):
            self.planet_detail.get_column(9+i).set_visible(minsummary)
        for i in range(8):
            self.planet_detail.get_column(12+i).set_visible(mindetails)

def cellformat_float(column, renderer, model, iter, user_data):
    """
    A function to format floats (or anything actually) using arbitrary format
    strings, for use inside a gtk.TreeViewColumn.
    """
    (colnumber, format) = user_data
    renderer.set_property('text', format % (model.get_value(iter, colnumber)))

# And now that we're way down here, start up the whole thing, assuming that
# we've been run from the commandline.
if __name__ == '__main__':
    uqmmap = UQMMap()
    uqmmap.run()
